"use strict";
var fs = require("fs");
var mkdirp = require("mkdirp");
var _path = require("path");
var rimraf = require("rimraf");
var pmb = require("packagemanager-backend");
var utils = require("./utils");
var tracker_1 = require("./tracker");
function createManager(options) {
    "use strict";
    if (options === void 0) { options = {}; }
    return new Manager(options)._setupBackend();
}
exports.createManager = createManager;
var Manager = (function () {
    function Manager(options) {
        if (options === void 0) { options = {}; }
        this.configPath = Manager.defaultConfigFile;
        this.offline = false;
        this.rootDir = Manager.defaultRootDir;
        this.repos = [];
        this.path = Manager.defaultPath;
        this.bundle = this.path + "/" + Manager.defaultBundleFile;
        options = utils.deepClone(options);
        this.configPath = options.configPath || this.configPath;
        this.repos = options.repos || this.repos;
        this.repos = this.repos.map(function (repo) { return repo; }); // copy
        if (this.repos.length === 0) {
            this.repos.push({
                url: Manager.defaultRepo,
                ref: Manager.defaultRef
            });
        }
        if (!this.repos[0].url) {
            this.repos[0].url = Manager.defaultRepo;
        }
        this.offline = options.offline || this.offline;
        this.tracker = new tracker_1.default();
        if (typeof options.insightOptout !== "undefined") {
            this.tracker.optOut = options.insightOptout;
        }
        if (this.tracker.optOut === false) {
            this.tracker.track();
        }
        this.savedRecipe = this._load();
        if (this.savedRecipe) {
            // for backward compatible
            var _recipe = this.savedRecipe;
            if (_recipe.baseRepo) {
                var baseRepo = _recipe.baseRepo;
                var baseRef = _recipe.baseRef;
                delete _recipe.baseRepo;
                delete _recipe.baseRef;
                this.savedRecipe.repos = this.savedRecipe.repos || [];
                this.savedRecipe.repos.unshift({
                    url: baseRepo,
                    ref: baseRef
                });
            }
            // for options
            if (this.repos.length === 0) {
                this.repos = this.savedRecipe.repos;
            }
            else if (options.repos && options.repos.length === 0) {
                this.repos = this.savedRecipe.repos;
            }
            this.rootDir = this.savedRecipe.rootDir || this.rootDir;
            this.path = this.savedRecipe.path || this.path;
        }
    }
    Manager.prototype._setupBackend = function () {
        var _this = this;
        return pmb.Manager
            .createManager({
            rootDir: this.rootDir,
            repos: this.repos
        })
            .then(function (backend) {
            _this.backend = backend;
            return _this;
        });
    };
    Manager.prototype._setupDefaultRecipe = function () {
        this.savedRecipe = this.savedRecipe || {};
        this.savedRecipe.repos = this.savedRecipe.repos || this.repos;
        this.savedRecipe.path = this.savedRecipe.path || this.path;
        this.savedRecipe.bundle = typeof this.savedRecipe.bundle !== "undefined" ? this.savedRecipe.bundle : this.bundle; // for this.recipe.bundle === null
        this.savedRecipe.link = this.savedRecipe.link || {};
        this.savedRecipe.link["npm"] = this.savedRecipe.link["npm"] || { include: true };
        this.savedRecipe.dependencies = this.savedRecipe.dependencies || {};
    };
    Manager.prototype.init = function (path) {
        if (path === void 0) { path = this.configPath; }
        this.tracker.track("init");
        this._setupDefaultRecipe();
        return this._save(path);
    };
    Manager.prototype._load = function (path) {
        if (path === void 0) { path = this.configPath; }
        if (fs.existsSync(path)) {
            var recipe = JSON.parse(fs.readFileSync(path, "utf8"));
            // backward compatible
            if (!recipe.repos || recipe.repos.length === 0) {
                recipe.repos = [{
                        url: recipe.baseRepo,
                        ref: recipe.baseRef
                    }];
            }
            delete recipe.baseRepo;
            delete recipe.baseRef;
            return recipe;
        }
        else {
            return null;
        }
    };
    Manager.prototype._save = function (path, recipe) {
        if (path === void 0) { path = this.configPath; }
        if (recipe === void 0) { recipe = this.savedRecipe; }
        var jsonContent = JSON.stringify(recipe, null, 2);
        mkdirp.sync(_path.resolve(path, "../"));
        fs.writeFileSync(path, jsonContent);
        return jsonContent;
    };
    Manager.prototype.search = function (phrase) {
        var _this = this;
        this.tracker.track("search", phrase);
        var promises;
        if (this.offline) {
            promises = [Promise.resolve(null)];
        }
        else {
            promises = this.backend.repos.map(function (repo) {
                return _this._fetchIfOutdated(repo);
            });
        }
        return Promise.all(promises)
            .then(function () {
            return _this.backend
                .search({
                globPatterns: [
                    "**/*.d.ts",
                    "!_infrastructure/**/*"
                ]
            });
        })
            .then(function (resultList) { return resultList.filter(function (result) { return result.fileInfo.path.indexOf(phrase) !== -1; }); })
            .then(function (resultList) { return _this._addWeightingAndSort(phrase, resultList, function (v) { return v.fileInfo.path; }).map(function (data) { return data.result; }); });
    };
    Manager.prototype.install = function (opts, phrases) {
        var _this = this;
        if (phrases) {
            phrases.forEach(function (phrase) {
                _this.tracker.track("install", phrase);
            });
        }
        if (!this.configPath) {
            return Promise.reject("configPath is required");
        }
        if (!fs.existsSync(this.configPath) && opts.save) {
            return Promise.reject(this.configPath + " is not exists");
        }
        this._setupDefaultRecipe();
        var promises = phrases.map(function (phrase) {
            return _this.search(phrase).then(function (resultList) {
                if (resultList.length === 1) {
                    return Promise.resolve(resultList[0]);
                }
                else if (resultList.length === 0) {
                    return Promise.reject(phrase + " is not found");
                }
                else {
                    var fileInfoWithWeight = _this._addWeightingAndSort(phrase, resultList, function (v) { return v.fileInfo.path; })[0];
                    if (fileInfoWithWeight && fileInfoWithWeight.weight === 1) {
                        return Promise.resolve(fileInfoWithWeight.result);
                    }
                    else {
                        return Promise.reject(phrase + " could not be identified. found: " + resultList.length);
                    }
                }
            });
        });
        return Promise.all(promises)
            .then(function (resultList) {
            if (!opts.save || opts.dryRun) {
                return resultList;
            }
            resultList.forEach(function (result) {
                var depName = result.fileInfo.path;
                if (_this.savedRecipe.dependencies[depName]) {
                    return;
                }
                _this.savedRecipe.dependencies[depName] = {
                    repo: result.repo.spec.url,
                    ref: result.fileInfo.ref
                };
                if (_this.savedRecipe) {
                    if (_this.savedRecipe.repos && _this.savedRecipe.repos[0].url === _this.savedRecipe.dependencies[depName].repo) {
                        delete _this.savedRecipe.dependencies[depName].repo;
                    }
                }
            });
            _this._save();
            return resultList;
        })
            .then(function (resultList) {
            var diff = {
                repos: _this.savedRecipe.repos,
                path: _this.savedRecipe.path,
                bundle: _this.savedRecipe.bundle,
                dependencies: {}
            };
            resultList.forEach(function (result) {
                var depName = result.fileInfo.path;
                diff.dependencies[depName] = _this.savedRecipe.dependencies[depName] || {
                    repo: result.repo.spec.url,
                    ref: result.fileInfo.ref
                };
            });
            return _this._installFromOptions(diff, opts);
        });
    };
    Manager.prototype.installFromFile = function (opts, retryWithFetch) {
        var _this = this;
        if (opts === void 0) { opts = {}; }
        if (retryWithFetch === void 0) { retryWithFetch = true; }
        if (retryWithFetch) {
            // installFromFile exec recursive, call tracker only once.
            this.tracker.track("installFromFile");
        }
        if (!this.configPath) {
            return Promise.reject("configPath is required");
        }
        var content = this._load();
        if (!content) {
            return Promise.reject(this.configPath + " is not exists");
        }
        return this._installFromOptions(content, opts)
            .catch(function (err) {
            // error raised when specified ref is not fetched yet.
            if (retryWithFetch) {
                return _this.fetch().then(function () { return _this.installFromFile(opts, false); });
            }
            else {
                return Promise.reject(err);
            }
        });
    };
    Manager.prototype._installFromOptions = function (recipe, opts) {
        var _this = this;
        if (opts === void 0) { opts = {}; }
        var baseRepo = recipe && recipe.repos && recipe.repos[0] || this.repos[0];
        return this.backend
            .getByRecipe({
            baseRepo: baseRepo.url,
            baseRef: baseRepo.ref,
            path: recipe.path,
            dependencies: recipe.dependencies,
            postProcessForDependency: function (result, depResult, content) {
                var dependencies = utils.extractDependencies(content.toString("utf8"));
                dependencies.forEach(function (detected) {
                    var obj = result.toDepNameAndPath(detected);
                    result.pushAdditionalDependency(obj.depName, {
                        repo: depResult.repo,
                        ref: depResult.ref,
                        path: obj.path
                    });
                });
            },
            resolveMissingDependency: function (result, missing) {
                if (missing.depth === 1) {
                    return null;
                }
                // first, from current process
                var newDep = result.dependencies[missing.depName];
                if (newDep) {
                    return Promise.resolve(newDep);
                }
                // second, from dtsm.json. can't use this.repos in `dtsm --remote .... install ....` context.
                if (_this.savedRecipe) {
                    newDep = _this.savedRecipe.dependencies[missing.depName];
                    var repo = _this.savedRecipe.repos[0];
                    if (newDep && repo) {
                        newDep.repo = newDep.repo || repo.url || Manager.defaultRepo;
                        newDep.ref = newDep.ref || repo.ref || Manager.defaultRef;
                    }
                    return Promise.resolve(newDep);
                }
                return Promise.resolve(null);
            }
        }).then(function (result) {
            var errors = result.dependenciesList.filter(function (depResult) { return !!depResult.error; });
            if (errors.length !== 0) {
                // TODO toString
                return Promise.reject(errors);
            }
            if (opts.dryRun) {
                return result;
            }
            // create definition files and bundle file
            result.dependenciesList.forEach(function (depResult) {
                _this._writeDefinitionFile(recipe, depResult);
                if (!recipe.bundle) {
                    return;
                }
                var depPath = _path.join(_path.dirname(_this.configPath), _this.path, depResult.depName);
                _this._addReferenceToBundle(recipe, depPath);
            });
            return result;
        });
    };
    Manager.prototype._addWeightingAndSort = function (phrase, list, getPath) {
        // TODO add something awesome weighing algorithm.
        return list
            .map(function (value) {
            var path = getPath(value);
            // exact match
            if (path === phrase) {
                return {
                    weight: 1,
                    result: value
                };
            }
            // library name match
            if (path === phrase + "/" + phrase + ".d.ts") {
                return {
                    weight: 1,
                    result: value
                };
            }
            // .d.t.s file match
            if (path.indexOf("/" + phrase + ".d.ts") !== -1) {
                return {
                    weight: 0.9,
                    result: value
                };
            }
            // directory name match
            if (path.indexOf(phrase + "/") === 0) {
                return {
                    weight: 0.8,
                    result: value
                };
            }
            // junk
            return {
                weight: 0.0,
                result: value
            };
        })
            .sort(function (a, b) { return b.weight - a.weight; });
    };
    Manager.prototype.update = function (opts) {
        var _this = this;
        this.tracker.track("update");
        if (!this.configPath) {
            return Promise.reject("configPath is required");
        }
        this._setupDefaultRecipe();
        // reset ref settings
        Object.keys(this.savedRecipe.dependencies).map(function (depName) {
            _this.savedRecipe.dependencies[depName].ref = null;
        });
        return this._installFromOptions(this.savedRecipe, opts)
            .then(function (result) {
            if (opts.dryRun) {
                return result;
            }
            Object.keys(result.dependencies).forEach(function (depName) {
                var depResult = result.dependencies[depName];
                _this._writeDefinitionFile(_this.savedRecipe, depResult);
                if (_this.savedRecipe.dependencies[depName]) {
                    _this.savedRecipe.dependencies[depName].ref = depResult.fileInfo.ref;
                }
            });
            if (opts.save) {
                _this._save();
            }
            return result;
        });
    };
    Manager.prototype.uninstall = function (opts, phrases) {
        var _this = this;
        this.tracker.track("uninstall");
        return this.installFromFile({ dryRun: true }, false)
            .then(function (result) {
            var topLevelDeps = Object.keys(result.dependencies).map(function (depName) { return result.dependencies[depName]; });
            if (topLevelDeps.length === 0) {
                return Promise.reject("this project doesn't have a dependency tree.");
            }
            var promises = phrases.map(function (phrase) {
                var weights = _this._addWeightingAndSort(phrase, topLevelDeps, function (dep) { return dep.depName; });
                weights = weights.filter(function (w) { return w.weight !== 0; });
                if (weights.length === 0) {
                    return Promise.reject(phrase + " is not found in config file");
                }
                if (weights[0].weight !== 1) {
                    return Promise.reject(phrase + " could not be identified. found: " + weights.length);
                }
                return Promise.resolve(weights[0].result);
            });
            return Promise.all(promises)
                .then(function (depList) {
                var removeList = [];
                var addToRemoveList = function (dep) {
                    if (!dep) {
                        return;
                    }
                    if (removeList.filter(function (rmDep) { return rmDep.depName === dep.depName; }).length !== 0) {
                        return;
                    }
                    removeList.push(dep);
                    if (!dep.dependencies) {
                        return;
                    }
                    Object.keys(dep.dependencies).forEach(function (depName) {
                        addToRemoveList(dep.dependencies[depName]);
                    });
                };
                depList.forEach(function (dep) {
                    // always remove from top level.
                    delete _this.savedRecipe.dependencies[dep.depName];
                    delete result.dependencies[dep.depName];
                    addToRemoveList(dep);
                });
                if (opts.dryRun) {
                    return depList;
                }
                removeList.forEach(function (dep) {
                    var unused = result.dependenciesList.every(function (exDep) { return exDep.depName !== dep.depName; });
                    if (unused) {
                        _this._removeDefinitionFile(_this.savedRecipe, dep);
                    }
                });
                if (!opts.save) {
                    return depList;
                }
                _this._save();
                return depList;
            });
        });
    };
    Manager.prototype.link = function (opts) {
        var _this = this;
        this.tracker.track("link");
        if (!this.configPath) {
            return Promise.reject("configPath is required");
        }
        this._setupDefaultRecipe();
        if (!this.savedRecipe.bundle) {
            return Promise.reject("bundle is required");
        }
        var linkInfo = this.savedRecipe.link || {};
        var resultList = [];
        {
            var depInfo = this._processLink(linkInfo["npm"], "npm", "package.json", "node_modules");
            linkInfo["npm"] = depInfo.link;
            if (!opts || !opts.dryRun) {
                depInfo.results.forEach(function (r) {
                    r.files.forEach(function (filePath) {
                        _this._addReferenceToBundle(_this.savedRecipe, filePath);
                    });
                });
            }
            resultList = resultList.concat(depInfo.results);
        }
        {
            var depInfo = this._processLink(linkInfo["bower"], "bower", "bower.json", "bower_components");
            linkInfo["bower"] = depInfo.link;
            if (!opts || !opts.dryRun) {
                depInfo.results.forEach(function (r) {
                    r.files.forEach(function (filePath) {
                        _this._addReferenceToBundle(_this.savedRecipe, filePath);
                    });
                });
            }
            resultList = resultList.concat(depInfo.results);
        }
        this.savedRecipe.link = linkInfo;
        if (opts && opts.save) {
            this._save();
        }
        return Promise.resolve(resultList);
    };
    Manager.prototype._processLink = function (linkInfo, managerName, configFileName, moduleDir) {
        if (linkInfo == null) {
        }
        else if (!linkInfo.include) {
            return { link: linkInfo, results: [] };
        }
        linkInfo = linkInfo || { include: true };
        var definitionList = [];
        var configPath = _path.join(_path.dirname(this.configPath), linkInfo.configPath || configFileName);
        if (!fs.existsSync(configPath)) {
            return { link: linkInfo, results: [] };
        }
        var configData = JSON.parse(fs.readFileSync(configPath, { encoding: "utf8" }));
        ["dependencies", "devDependencies"].forEach(function (propName) {
            Object.keys(configData[propName] || {}).forEach(function (depName) {
                var depConfigPath = _path.join(_path.dirname(configPath), moduleDir, depName, configFileName);
                if (!fs.existsSync(depConfigPath)) {
                    return;
                }
                var depConfig = JSON.parse(fs.readFileSync(depConfigPath, { encoding: "utf8" }));
                var definitionInfo = depConfig["typescript"];
                if (!definitionInfo) {
                    return;
                }
                var resultList = [];
                ["definition", "definitions"].forEach(function (propName) {
                    if (typeof definitionInfo[propName] === "string") {
                        resultList.push(definitionInfo[propName]);
                    }
                    else if (Array.isArray(definitionInfo[propName])) {
                        resultList = resultList.concat(definitionInfo[propName]);
                    }
                });
                definitionList.push({
                    managerName: managerName,
                    depName: depName,
                    files: resultList.map(function (filePath) { return _path.join(_path.dirname(configPath), moduleDir, depName, filePath); })
                });
            });
        });
        return { link: linkInfo, results: definitionList };
    };
    Manager.prototype._writeDefinitionFile = function (recipe, depResult) {
        var path = _path.resolve(_path.dirname(this.configPath), recipe.path, depResult.depName);
        mkdirp.sync(_path.resolve(path, "../"));
        fs.writeFileSync(path, depResult.content.toString("utf8"));
    };
    Manager.prototype._removeDefinitionFile = function (recipe, depResult) {
        var path = _path.resolve(_path.dirname(this.configPath), recipe.path, depResult.depName);
        try {
            fs.unlinkSync(path);
            var contents = fs.readdirSync(_path.dirname(path));
            if (contents.length === 0) {
                rimraf.sync(_path.dirname(path));
            }
        }
        catch (e) {
        }
    };
    Manager.prototype._addReferenceToBundle = function (recipe, pathFromCwd) {
        var bundleContent = "";
        var bundlePath = _path.join(_path.dirname(this.configPath), recipe.bundle);
        if (fs.existsSync(bundlePath)) {
            bundleContent = fs.readFileSync(bundlePath, "utf8");
        }
        else {
            mkdirp.sync(_path.dirname(bundlePath));
        }
        var referencePath = _path.relative(_path.dirname(bundlePath), pathFromCwd);
        if (_path.posix) {
            referencePath = referencePath.replace(new RegExp("\\" + _path.win32.sep, "g"), _path.posix.sep);
        }
        var referenceComment = ("/// <reference path=\"" + referencePath + "\" />") + "\n";
        if (bundleContent.indexOf(referenceComment) === -1) {
            fs.appendFileSync(bundlePath, referenceComment, { encoding: "utf8" });
        }
    };
    Manager.prototype.refs = function () {
        var _this = this;
        var promises;
        if (this.offline) {
            promises = [Promise.resolve(null)];
        }
        else {
            promises = this.backend.repos.map(function (repo) {
                return _this._fetchIfOutdated(repo);
            });
        }
        return Promise.all(promises)
            .then(function () {
            var repo = _this.backend.repos[0];
            return repo.open().then(function (fs) { return fs.showRef(); });
        });
    };
    Manager.prototype.fetch = function () {
        var _this = this;
        this.tracker.track("fetch");
        var promises = this.backend.repos.map(function (repo) {
            return _this._fetchRepo(repo);
        });
        return Promise.all(promises).then(function () { return null; });
    };
    Manager.prototype._fetchIfOutdated = function (repo) {
        if (this._checkOutdated(repo.spec.url)) {
            return this._fetchRepo(repo);
        }
        else {
            return Promise.resolve(repo);
        }
    };
    Manager.prototype._fetchRepo = function (repo) {
        var _this = this;
        console.log("fetching " + repo.spec.url);
        return Promise
            .resolve(null)
            .then(function () { return repo.fetchAll(); })
            .then(function () {
            _this._setLastFetchAt(repo.spec.url);
            return repo;
        });
    };
    Manager.prototype._checkOutdated = function (repoUrl) {
        var fetchAt = this._getLastFetchAt(repoUrl);
        // 15min
        return fetchAt + 15 * 60 * 1000 < Date.now();
    };
    Manager.prototype._getLastFetchAt = function (repoID) {
        var config = this.backend.loadConfig() || {};
        config.repositories = config.repositories || {};
        var repo = config.repositories[repoID] || { fetchAt: null };
        return repo.fetchAt;
    };
    Manager.prototype._setLastFetchAt = function (repoID, fetchAt) {
        if (fetchAt === void 0) { fetchAt = Date.now(); }
        var config = this.backend.loadConfig() || {};
        config.repositories = config.repositories || {};
        config.repositories[repoID] = {
            fetchAt: fetchAt
        };
        this.backend.saveConfig(config);
    };
    Manager.defaultConfigFile = process.cwd() + "/dtsm.json";
    Manager.defaultRootDir = "~/.dtsm";
    Manager.defaultRepo = "https://github.com/DefinitelyTyped/DefinitelyTyped.git";
    Manager.defaultRef = "master";
    Manager.defaultPath = "typings";
    Manager.defaultBundleFile = "bundle.d.ts";
    return Manager;
})();
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = Manager;
//# sourceMappingURL=manager.js.map